# -*- coding: utf-8 -*-
#
# This file is part of Invenio.
# Copyright (C) 2017 CERN.
#
# Invenio is free software; you can redistribute it
# and/or modify it under the terms of the GNU General Public License as
# published by the Free Software Foundation; either version 2 of the
# License, or (at your option) any later version.
#
# Invenio is distributed in the hope that it will be
# useful, but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the GNU
# General Public License for more details.
#
# You should have received a copy of the GNU General Public License
# along with Invenio; if not, write to the
# Free Software Foundation, Inc., 59 Temple Place, Suite 330, Boston,
# MA 02111-1307, USA.
#
# In applying this license, CERN does not
# waive the privileges and immunities granted to it by virtue of its status
# as an Intergovernmental Organization or submit itself to any jurisdiction.

"""Minimal Flask application example.

First install Invenio-PIDRelations, setup the application and load
fixture data by running:

.. code-block:: console

   $ pip install -e .[all]
   $ cd examples
   $ ./app-setup.sh
   $ ./app-fixtures.sh

Next, start the development server:

.. code-block:: console

   $ export FLASK_APP=app.py FLASK_DEBUG=1
   $ flask run

and open the example application in your browser:

.. code-block:: console

    $ open http://127.0.0.1:5000/

To reset the example application run:

.. code-block:: console

    $ ./app-teardown.sh
"""

from __future__ import absolute_import, print_function

from flask import Flask, redirect, render_template, request, url_for
from invenio_db import InvenioDB, db
from invenio_indexer import InvenioIndexer
from invenio_pidstore import InvenioPIDStore
from invenio_pidstore.models import PersistentIdentifier, PIDStatus
from invenio_pidstore.providers.recordid import RecordIdProvider
from invenio_pidstore.resolver import Resolver
from invenio_records import InvenioRecords
from invenio_records.api import Record
from marshmallow import Schema, fields

from invenio_pidrelations import InvenioPIDRelations
from invenio_pidrelations.contrib.versioning import PIDVersioning, \
    versioning_blueprint
from invenio_pidrelations.models import PIDRelation
from invenio_pidrelations.serializers.schemas import PIDRelationsMixin
from invenio_pidrelations.utils import resolve_relation_type_config

# Create Flask application
app = Flask(__name__, template_folder='.')
app.config['TEMPLATES_AUTO_RELOAD'] = True

InvenioDB(app)
InvenioPIDStore(app)
InvenioPIDRelations(app)
app.register_blueprint(versioning_blueprint)
InvenioIndexer(app)
InvenioRecords(app)

record_resolver = Resolver(
    pid_type='recid', object_type='rec', getter=Record.get_record
)


class SimpleRecordSchema(Schema, PIDRelationsMixin):
    """Tiny schema for our simple record."""

    recid = fields.Str()
    title = fields.Str()
    body = fields.Str()


@app.route('/')
def index():
    relation_id = resolve_relation_type_config('version').id
    heads = (
        PersistentIdentifier.query
        .join(
            PIDRelation,
            PIDRelation.parent_id == PersistentIdentifier.id)
        .filter(
            PIDRelation.relation_type == relation_id)
        .distinct())
    return render_template('index.html', heads=heads)


@app.route('/create', methods=['POST'])
def create():
    create_simple_record(request.form)
    return redirect(url_for('index'))


@app.template_filter()
def to_record(pid):
    schema = SimpleRecordSchema()
    schema.context = dict(pid=pid)
    rec = schema.dump(record_resolver.resolve(pid.pid_value)[1])
    return rec.data


def create_simple_record(data):
    # Create the record and mint a PID
    metadata, errors = SimpleRecordSchema().load(data)
    parent = data.get('parent')
    if parent != 'new':
        metadata['conceptrecid'] = parent

    rec = Record.create(metadata)

    record_minter(rec.id, rec)
    db.session.commit()


def record_minter(record_uuid, data):
    parent = data.get('conceptrecid')
    if not parent:
        parent_pid = RecordIdProvider.create(object_type='rec',
                                             object_uuid=None,
                                             status=PIDStatus.REGISTERED).pid
        data['conceptrecid'] = parent_pid.pid_value
    else:
        parent_pid = PersistentIdentifier.get(
            pid_type=RecordIdProvider.pid_type, pid_value=parent)
    provider = RecordIdProvider.create('rec', record_uuid)
    data['recid'] = provider.pid.pid_value

    versioning = PIDVersioning(parent=parent_pid)
    versioning.insert_child(child=provider.pid)
    return provider.pid



# -*- coding: utf-8 -*-
#
# This file is part of Invenio.
# Copyright (C) 2017 CERN.
#
# Invenio is free software; you can redistribute it
# and/or modify it under the terms of the GNU General Public License as
# published by the Free Software Foundation; either version 2 of the
# License, or (at your option) any later version.
#
# Invenio is distributed in the hope that it will be
# useful, but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the GNU
# General Public License for more details.
#
# You should have received a copy of the GNU General Public License
# along with Invenio; if not, write to the
# Free Software Foundation, Inc., 59 Temple Place, Suite 330, Boston,
# MA 02111-1307, USA.
#
# In applying this license, CERN does not
# waive the privileges and immunities granted to it by virtue of its status
# as an Intergovernmental Organization or submit itself to any jurisdiction.

"""Schema tests."""

from marshmallow import Schema

from invenio_pidrelations.models import PIDRelation
from invenio_pidrelations.serializers.schemas import PIDRelationsMixin, \
    RelationSchema


class SampleRecordSchema(Schema, PIDRelationsMixin):
    """Sample record schema."""
    pass


def test_schema(app, nested_pids_and_relations):
    """Test the marshmallow schema serialization."""
    schema = SampleRecordSchema(strict=True)

    pids, exp_relations = nested_pids_and_relations
    for p_idx in exp_relations.keys():
        pid = pids[p_idx]
        expected = exp_relations[p_idx]
        input_data = {'pid': pid}
        schema.context['pid'] = pid
        data, errors = schema.dump(input_data)
        assert not errors
        assert expected == data  # Test against hand-crafted fixture
    pass


def test_custom_schema(app, nested_pids_and_relations, custom_relation_schema):
    """Test the marshmallow schema serialization with custom schema."""
    schema = SampleRecordSchema(strict=True)
    pids, exp_relations = nested_pids_and_relations

    pid = pids[4]
    input_data = {'pid': pid}
    schema.context['pid'] = pid
    data, errors = schema.dump(input_data)
    expected = {
        'relations': {
            'version': [
                {
                    'children': [{'pid_type': 'recid', 'pid_value': '2'},
                                 {'pid_type': 'recid', 'pid_value': '3'},
                                 {'pid_type': 'recid', 'pid_value': '4'}],
                    'has_three_children': True,
                },
            ],
            'ordered': [
                {
                    'children': [{'pid_type': 'recid', 'pid_value': '6'},
                                 {'pid_type': 'recid', 'pid_value': '4'},
                                 {'pid_type': 'recid', 'pid_value': '7'}],
                    'has_three_children': True,
                },
                {
                    'children': [{'pid_type': 'recid', 'pid_value': '8'},
                                 {'pid_type': 'recid', 'pid_value': '9'}],
                    'has_three_children': False,
                },
            ],
            'unordered': [
                {
                    'children': [{'pid_type': 'recid', 'pid_value': '4'},
                                 {'pid_type': 'recid', 'pid_value': '11'}],
                    'has_three_children': False,
                },
            ],
        }
    }
    assert not errors
    assert expected == data



# -*- coding: utf-8 -*-
#
# This file is part of Invenio.
# Copyright (C) 2017 CERN.
#
# Invenio is free software; you can redistribute it
# and/or modify it under the terms of the GNU General Public License as
# published by the Free Software Foundation; either version 2 of the
# License, or (at your option) any later version.
#
# Invenio is distributed in the hope that it will be
# useful, but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the GNU
# General Public License for more details.
#
# You should have received a copy of the GNU General Public License
# along with Invenio; if not, write to the
# Free Software Foundation, Inc., 59 Temple Place, Suite 330, Boston,
# MA 02111-1307, USA.
#
# In applying this license, CERN does not
# waive the privileges and immunities granted to it by virtue of its status
# as an Intergovernmental Organization or submit itself to any jurisdiction.

"""Pytest configuration."""

from __future__ import absolute_import, print_function

import os
import shutil
import tempfile

import pytest
from elasticsearch.exceptions import RequestError
from flask import Flask
from flask_babelex import Babel
from invenio_db import db as db_
from invenio_db import InvenioDB
from invenio_indexer import InvenioIndexer
from invenio_indexer.api import RecordIndexer
from invenio_pidstore import InvenioPIDStore
from invenio_pidstore.models import PersistentIdentifier, PIDStatus
from invenio_records import InvenioRecords, Record
from invenio_search import InvenioSearch, current_search, current_search_client
from marshmallow import fields
from sqlalchemy_utils.functions import create_database, database_exists

from invenio_pidrelations import InvenioPIDRelations
from invenio_pidrelations.config import RelationType
from invenio_pidrelations.contrib.versioning import PIDVersioning
from invenio_pidrelations.models import PIDRelation
from invenio_pidrelations.serializers.schemas import RelationSchema
from invenio_pidrelations.utils import resolve_relation_type_config


@pytest.yield_fixture()
def instance_path():
    """Temporary instance path."""
    path = tempfile.mkdtemp()
    yield path
    shutil.rmtree(path)


@pytest.fixture()
def base_app(instance_path):
    """Flask application fixture."""
    app_ = Flask('testapp', instance_path=instance_path)
    app_.config.update(
        SECRET_KEY='SECRET_KEY',
        SQLALCHEMY_DATABASE_URI=os.environ.get(
            'SQLALCHEMY_DATABASE_URI', 'sqlite:///test.db'),
        SQLALCHEMY_TRACK_MODIFICATIONS=True,
        TESTING=True,
    )
    InvenioPIDStore(app_)
    InvenioDB(app_)
    InvenioPIDRelations(app_)
    InvenioRecords(app_)
    InvenioIndexer(app_)
    InvenioSearch(app_)
    Babel(app_)
    return app_


@pytest.yield_fixture()
def app(base_app):
    """Flask application fixture."""
    with base_app.app_context():
        yield base_app


@pytest.yield_fixture()
def es(app):
    """Provide elasticsearch access."""
    try:
        list(current_search.create())
    except RequestError:
        list(current_search.delete(ignore=[400, 404]))
        list(current_search.create())
    current_search_client.indices.refresh()
    yield current_search_client
    list(current_search.delete(ignore=[404]))


@pytest.yield_fixture()
def db(app):
    """Database fixture."""
    if not database_exists(str(db_.engine.url)):
        create_database(str(db_.engine.url))
    db_.create_all()
    yield db_
    db_.session.remove()
    db_.drop_all()


@pytest.fixture()
def pids(app, db):
    """Test PIDs fixture."""
    # TODO: Head PIDs do not have redirects as they are created outside API
    h1 = PersistentIdentifier.create('recid', 'foobar', object_type='rec',
                                     status=PIDStatus.REGISTERED)
    h1v1 = PersistentIdentifier.create('recid', 'foobar.v1', object_type='rec')
    h1v2 = PersistentIdentifier.create('recid', 'foobar.v2', object_type='rec')
    h1v3 = PersistentIdentifier.create('recid', 'foobar.v3', object_type='rec')

    ORDERED = resolve_relation_type_config('ordered').id
    UNORDERED = resolve_relation_type_config('unordered').id
    PIDRelation.create(h1, h1v1, ORDERED, 0)
    PIDRelation.create(h1, h1v2, ORDERED, 1)
    PIDRelation.create(h1, h1v3, ORDERED, 2)
    h1.redirect(h1v3)

    h2 = PersistentIdentifier.create('recid', 'spam', object_type='rec',
                                     status=PIDStatus.REGISTERED)
    h2v1 = PersistentIdentifier.create('recid', 'spam.v1')
    PIDRelation.create(h2, h2v1, ORDERED, 0)
    h2.redirect(h2v1)

    c1 = PersistentIdentifier.create('recid', 'bazbar')
    c1r1 = PersistentIdentifier.create('recid', 'resource1')
    c1r2 = PersistentIdentifier.create('recid', 'resource2')

    pid1 = PersistentIdentifier.create('recid', 'eggs')
    PIDRelation.create(c1, c1r1, UNORDERED, None)
    PIDRelation.create(c1, c1r2, UNORDERED, None)
    return {
        'h1': h1,
        'h1v1': h1v1,
        'h1v2': h1v2,
        'h1v3': h1v3,
        'h2': h2,
        'h2v1': h2v1,
        'c1': c1,
        'c1r1': c1r1,
        'c1r2': c1r2,
        'pid1': pid1,
    }


@pytest.fixture()
def nested_pids_and_relations(app, db):
    """Fixture for a nested PIDs and the expected serialized relations."""
    # Create some PIDs and connect them into different nested PID relations
    pids = {}
    for idx in range(1, 12):
        pid_value = str(idx)
        p = PersistentIdentifier.create('recid', pid_value, object_type='rec',
                                        status=PIDStatus.REGISTERED)
        pids[idx] = p

    ORDERED = resolve_relation_type_config('ordered').id
    UNORDERED = resolve_relation_type_config('unordered').id
    VERSION = resolve_relation_type_config('version').id

    #    1  (Version)
    #  / | \
    # 2  3 4
    PIDRelation.create(pids[1], pids[2], VERSION, 0)
    PIDRelation.create(pids[1], pids[3], VERSION, 1)
    PIDRelation.create(pids[1], pids[4], VERSION, 2)

    #    5  (Ordered)
    #  / | \
    # 6  4 7
    PIDRelation.create(pids[5], pids[6], ORDERED, 0)
    PIDRelation.create(pids[5], pids[4], ORDERED, 1)
    PIDRelation.create(pids[5], pids[7], ORDERED, 2)

    #    4  (Ordered)
    #  / |
    # 8  9
    PIDRelation.create(pids[4], pids[8], ORDERED, 0)
    PIDRelation.create(pids[4], pids[9], ORDERED, 1)

    #   10  (Unordered)
    #  / |
    # 4  11
    PIDRelation.create(pids[10], pids[4], UNORDERED, None)
    PIDRelation.create(pids[10], pids[11], UNORDERED, None)

    # Define the expected PID relation tree for of the PIDs
    expected_relations = {}
    expected_relations[4] = {
        'relations': {
            'version': [
                {
                    'children': [{'pid_type': 'recid', 'pid_value': '2'},
                                 {'pid_type': 'recid', 'pid_value': '3'},
                                 {'pid_type': 'recid', 'pid_value': '4'}],
                    'is_child': True,
                    'index': 2,
                    'previous': {'pid_type': 'recid', 'pid_value': '3'},
                    'next': None,
                    'is_last': True,
                    'is_ordered': True,
                    'is_parent': False,
                    'parent': {'pid_type': 'recid', 'pid_value': '1'},
                    'type': 'version'
                }
            ],
            'ordered': [
                {
                    'children': [{'pid_type': 'recid', 'pid_value': '6'},
                                 {'pid_type': 'recid', 'pid_value': '4'},
                                 {'pid_type': 'recid', 'pid_value': '7'}],
                    'is_child': True,
                    'index': 1,
                    'previous': {'pid_type': 'recid', 'pid_value': '6'},
                    'next': {'pid_type': 'recid', 'pid_value': '7'},
                    'is_last': False,
                    'is_ordered': True,
                    'is_parent': False,
                    'parent': {'pid_type': 'recid', 'pid_value': '5'},
                    'type': 'ordered'
                },
                {
                    'children': [{'pid_type': 'recid', 'pid_value': '8'},
                                 {'pid_type': 'recid', 'pid_value': '9'}],
                    'is_child': False,
                    'index': None,
                    'previous': None,
                    'next': None,
                    'is_last': None,
                    'is_ordered': True,
                    'is_parent': True,
                    'parent': {'pid_type': 'recid', 'pid_value': '4'},
                    'type': 'ordered'
                }
            ],
            'unordered': [
                {
                    'children': [{'pid_type': 'recid', 'pid_value': '4'},
                                 {'pid_type': 'recid', 'pid_value': '11'}],
                    'is_child': True,
                    'index': None,
                    'previous': None,
                    'next': None,
                    'is_last': False,
                    'is_ordered': True,
                    'is_parent': False,
                    'parent': {'pid_type': 'recid', 'pid_value': '10'},
                    'type': 'unordered'
                },
            ]
        }
    }
    return pids, expected_relations


class CustomRelationSchema(RelationSchema):
    """PID relation Schema containing only children and one extra field."""

    class Meta:
        """Meta fields of the schema."""

        fields = ("children", "has_three_children", )

    has_three_children = fields.Method('dump_has_three_children')

    def dump_has_three_children(self, obj):
        """Dump information if relation has exactly three children."""
        return obj.children.count() == 3


@pytest.yield_fixture()
def custom_relation_schema(app):
    """Fixture for PID relations config with custom schemas."""
    orig = app.config['PIDRELATIONS_RELATION_TYPES']
    app.config['PIDRELATIONS_RELATION_TYPES'] = [
        RelationType(0, 'ordered', 'Ordered',
                     'invenio_pidrelations.api:PIDConceptOrdered',
                     CustomRelationSchema),
        RelationType(1, 'unordered', 'Unordered',
                     'invenio_pidrelations.api:PIDConcept',
                     CustomRelationSchema),
        RelationType(2, 'version', 'Version',
                     'invenio_pidrelations.contrib.versioning:PIDVersioning',
                     CustomRelationSchema),
    ]
    yield app
    app.config['PIDRELATIONS_RELATION_TYPES'] = orig


@pytest.fixture()
def version_pids(app, db):
    """Versioned PIDs fixture with one parent and two versions."""
    h1 = PersistentIdentifier.create('recid', 'foobar', object_type='rec',
                                     status=PIDStatus.REGISTERED)
    h1v1 = PersistentIdentifier.create('recid', 'foobar.v1', object_type='rec',
                                       status=PIDStatus.REGISTERED)
    h1v2 = PersistentIdentifier.create('recid', 'foobar.v2', object_type='rec',
                                       status=PIDStatus.REGISTERED)
    pv = PIDVersioning(parent=h1)
    pv.insert_child(h1v1)
    pv.insert_child(h1v2)
    db.session.commit()
    return {
        'h1': h1,
        'h1v1': h1v1,
        'h1v2': h1v2,
    }


@pytest.fixture()
def records(pids, db):
    """Fixture for the records."""
    pid_versions = ['h1v1', 'h1v2', 'h2v1']
    schema = {
        'type': 'object',
        'properties': {
            'title': {'type': 'string'},
        },
    }
    data = {
        name: {'title': 'Test version {}'.format(name),
               'recid': pids[name].pid_value,
               '$schema': schema}
        for name in pid_versions
    }
    records = dict()
    for name in pid_versions:
        record = Record.create(data[name])
        pids[name].assign('rec', record.id)
        records[name] = record
    return records


@pytest.fixture()
def indexed_records(es, records):
    """Fixture for the records, which are already indexed."""
    # es.indices.flush('*')
    # # delete all elasticsearch indices and recreate them
    # for deleted in current_search.delete(ignore=[404]):
    #     pass
    # for created in current_search.create(None):
    #     pass
    # flush the indices so that indexed records are searchable
    for pid_name, record in records.items():
        RecordIndexer().index(record)
    es.indices.flush('*')
    return records



# -*- coding: utf-8 -*-
#
# This file is part of Invenio.
# Copyright (C) 2014, 2015, 2016 CERN.
#
# Invenio is free software; you can redistribute it
# and/or modify it under the terms of the GNU General Public License as
# published by the Free Software Foundation; either version 2 of the
# License, or (at your option) any later version.
#
# Invenio is distributed in the hope that it will be
# useful, but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the GNU
# General Public License for more details.
#
# You should have received a copy of the GNU General Public License
# along with Invenio; if not, write to the
# Free Software Foundation, Inc., 59 Temple Place, Suite 330, Boston,
# MA 02111-1307, USA.
#
# In applying this license, CERN does not
# waive the privileges and immunities granted to it by virtue of its status
# as an Intergovernmental Organization or submit itself to any jurisdiction.

"""Persistent identifier's relations models."""

from __future__ import absolute_import, print_function

import logging

from flask_babelex import gettext
from invenio_db import db
from invenio_pidstore.models import PersistentIdentifier
from speaklater import make_lazy_gettext
from sqlalchemy.exc import IntegrityError
from sqlalchemy.orm import backref
from sqlalchemy_utils.models import Timestamp

_ = make_lazy_gettext(lambda: gettext)

logger = logging.getLogger('invenio-pidrelations')


class PIDRelation(db.Model, Timestamp):
    """Model persistent identifier relations."""

    __tablename__ = 'pidrelations_pidrelation'

    parent_id = db.Column(
        db.Integer,
        db.ForeignKey(PersistentIdentifier.id, onupdate="CASCADE",
                      ondelete="CASCADE",
                      name='fk_pidrelation_parent_pidstore_pid'),
        nullable=False,
        primary_key=True,
        )
    """Parent PID of the relation."""

    child_id = db.Column(
        db.Integer,
        db.ForeignKey(PersistentIdentifier.id, onupdate="CASCADE",
                      ondelete="CASCADE",
                      name='fk_pidrelation_parent_pidstore_pid'),
        nullable=False,
        primary_key=True)
    """Child PID of the relation."""

    relation_type = db.Column(
        db.SmallInteger(),
        nullable=False)
    """Type of relation between the parent and child PIDs."""

    index = db.Column(db.Integer, nullable=True)
    """Index of the PID relation (e.g.: modeling ordered sequence of PIDs)."""

    #
    # Relations
    #
    parent = db.relationship(
        PersistentIdentifier,
        primaryjoin=PersistentIdentifier.id == parent_id,
        backref=backref('child_relations', lazy='dynamic',
                        cascade='all,delete'))

    child = db.relationship(
        PersistentIdentifier,
        primaryjoin=PersistentIdentifier.id == child_id,
        backref=backref('parent_relations', lazy='dynamic',
                        cascade='all,delete'))

    def __repr__(self):
        """Text representation of a PID relation."""
        return "<PIDRelation: ({r.parent.pid_type}:{r.parent.pid_value}) -> " \
               "({r.child.pid_type}:{r.child.pid_value}) " \
               "(Type: {r.relation_type}, Idx: {r.index})>".format(r=self)

    @classmethod
    def get_parent_relations(cls, pid):
        """Get all relations where given PID is a parent."""
        return cls.query.filter((cls.parent == pid))

    @classmethod
    def get_child_relations(cls, pid):
        """Get all relations where given PID is a child."""
        return cls.query.filter((cls.child == pid))

    @classmethod
    def create(cls, parent, child, relation_type, index=None):
        """Create a PID relation for given parent and child."""
        try:
            with db.session.begin_nested():
                obj = cls(parent_id=parent.id,
                          child_id=child.id,
                          relation_type=relation_type,
                          index=index)
                db.session.add(obj)
        except IntegrityError:
            raise Exception("PID Relation already exists.")
            # msg = "PIDRelation already exists: " \
            #       "{0} -> {1} ({2})".format(
            #         parent_pid, child_pid, relation_type)
            # logger.exception(msg)
            # raise Exception(msg)
        return obj

    @classmethod
    def relation_exists(self, parent, child, relation_type):
        """Determine if given relation already exists."""
        return PIDRelation.query.filter_by(
            child_pid_id=child.id,
            parent_pid_id=parent.id,
            relation_type=relation_type).count() > 0


__all__ = (
    'PIDRelation',
)



# -*- coding: utf-8 -*-
#
# This file is part of Invenio.
# Copyright (C) 2017 CERN.
#
# Invenio is free software; you can redistribute it
# and/or modify it under the terms of the GNU General Public License as
# published by the Free Software Foundation; either version 2 of the
# License, or (at your option) any later version.
#
# Invenio is distributed in the hope that it will be
# useful, but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the GNU
# General Public License for more details.
#
# You should have received a copy of the GNU General Public License
# along with Invenio; if not, write to the
# Free Software Foundation, Inc., 59 Temple Place, Suite 330, Boston,
# MA 02111-1307, USA.
#
# In applying this license, CERN does not
# waive the privileges and immunities granted to it by virtue of its status
# as an Intergovernmental Organization or submit itself to any jurisdiction.

"""PID relations search filters."""

from __future__ import absolute_import, print_function

from elasticsearch_dsl.query import Bool, Q


class LatestVersionFilter(object):
    """Shortcut for defining default filters with query parser."""

    def __init__(self, query=None, query_parser=None):
        """Build filter property with query parser."""
        self._query = Q('term', **{'relation.version.is_latest': True})
        if query is not None:
            self._query = Bool(
                must=[
                    query,
                    self._query
                ],
            )
        self.query_parser = query_parser or (lambda x: x)

    @property
    def query(self):
        """Build lazy query if needed."""
        return self._query() if callable(self._query) else self._query

    def __get__(self, obj, objtype):
        """Return parsed query."""
        return self.query_parser(self.query)



# -*- coding: utf-8 -*-
#
# This file is part of Invenio.
# Copyright (C) 2017 CERN.
#
# Invenio is free software; you can redistribute it
# and/or modify it under the terms of the GNU General Public License as
# published by the Free Software Foundation; either version 2 of the
# License, or (at your option) any later version.
#
# Invenio is distributed in the hope that it will be
# useful, but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the GNU
# General Public License for more details.
#
# You should have received a copy of the GNU General Public License
# along with Invenio; if not, write to the
# Free Software Foundation, Inc., 59 Temple Place, Suite 330, Boston,
# MA 02111-1307, USA.
#
# In applying this license, CERN does not
# waive the privileges and immunities granted to it by virtue of its status
# as an Intergovernmental Organization or submit itself to any jurisdiction.

"""Records contribution module tests."""

from __future__ import absolute_import, print_function

import pytest
from invenio_pidstore.models import PersistentIdentifier

from invenio_pidrelations.contrib.records import RecordDraft
from invenio_pidrelations.models import PIDRelation
from invenio_pidrelations.utils import resolve_relation_type_config


def test_record_draft(app, db):
    """Test RecordDraft API."""

    assert PersistentIdentifier.query.count() == 0
    assert PIDRelation.query.count() == 0

    d1 = PersistentIdentifier.create('depid', '1', object_type='rec')
    r1 = PersistentIdentifier.create('recid', '1', object_type='rec')
    assert PersistentIdentifier.query.count() == 2

    RecordDraft.link(recid=r1, depid=d1)
    assert PIDRelation.query.count() == 1

    pr = PIDRelation.query.one()
    RECORD_DRAFT = resolve_relation_type_config('record_draft').id
    assert pr.relation_type == RECORD_DRAFT
    assert pr.index is None
    assert pr.parent == r1
    assert pr.child == d1

    d2 = PersistentIdentifier.create('depid', '2', object_type='rec')
    r2 = PersistentIdentifier.create('recid', '2', object_type='rec')

    with pytest.raises(Exception) as excinfo:
        RecordDraft.link(recid=r1, depid=d2)
    assert 'already has a depid as a draft' in str(excinfo.value)

    with pytest.raises(Exception) as excinfo:
        RecordDraft.link(recid=r2, depid=d1)
    assert 'already is a draft of a recid' in str(excinfo.value)



# -*- coding: utf-8 -*-
#
# This file is part of Invenio.
# Copyright (C) 2017 CERN.
#
# Invenio is free software; you can redistribute it
# and/or modify it under the terms of the GNU General Public License as
# published by the Free Software Foundation; either version 2 of the
# License, or (at your option) any later version.
#
# Invenio is distributed in the hope that it will be
# useful, but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the GNU
# General Public License for more details.
#
# You should have received a copy of the GNU General Public License
# along with Invenio; if not, write to the
# Free Software Foundation, Inc., 59 Temple Place, Suite 330, Boston,
# MA 02111-1307, USA.
#
# In applying this license, CERN does not
# waive the privileges and immunities granted to it by virtue of its status
# as an Intergovernmental Organization or submit itself to any jurisdiction.

"""PID relations utility functions."""

from __future__ import absolute_import, print_function

import six
from flask import current_app
from werkzeug.utils import import_string


def obj_or_import_string(value, default=None):
    """Import string or return object.

    :params value: Import path or class object to instantiate.
    :params default: Default object to return if the import fails.
    :returns: The imported object.
    """
    if isinstance(value, six.string_types):
        return import_string(value)
    elif value:
        return value
    return default


def resolve_relation_type_config(value):
    """Resolve the relation type to config object.

    Resolve relation type from string (e.g.:  serialization) or int (db value)
    to the full config object.
    """
    relation_types = current_app.config['PIDRELATIONS_RELATION_TYPES']
    if isinstance(value, six.string_types):
        try:
            obj = next(rt for rt in relation_types if rt.name == value)
        except StopIteration as e:
            raise ValueError("Relation name '{0}' is not configured.".format(
                value))

    elif isinstance(value, int):
        try:
            obj = next(rt for rt in relation_types if rt.id == value)
        except StopIteration as e:
            raise ValueError("Relation ID {0} is not configured.".format(
                value))
    else:
        raise ValueError("Type of value '{0}' is not supported for resolving.")
    api_class = obj_or_import_string(obj.api)
    schema_class = obj_or_import_string(obj.schema)
    return obj.__class__(obj.id, obj.name, obj.label, api_class, schema_class)



# -*- coding: utf-8 -*-
#
# This file is part of Invenio.
# Copyright (C) 2017 CERN.
#
# Invenio is free software; you can redistribute it
# and/or modify it under the terms of the GNU General Public License as
# published by the Free Software Foundation; either version 2 of the
# License, or (at your option) any later version.
#
# Invenio is distributed in the hope that it will be
# useful, but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the GNU
# General Public License for more details.
#
# You should have received a copy of the GNU General Public License
# along with Invenio; if not, write to the
# Free Software Foundation, Inc., 59 Temple Place, Suite 330, Boston,
# MA 02111-1307, USA.
#
# In applying this license, CERN does not
# waive the privileges and immunities granted to it by virtue of its status
# as an Intergovernmental Organization or submit itself to any jurisdiction.

"""PID relations indexers."""

from __future__ import absolute_import, print_function

from invenio_pidstore.models import PersistentIdentifier

from .proxies import current_pidrelations
from .serializers.utils import serialize_relations


def index_relations(sender, json=None, record=None, index=None, **kwargs):
    """Add relations to the indexed record."""
    pid = PersistentIdentifier.query.filter(
        PersistentIdentifier.object_uuid == record.id,
        PersistentIdentifier.pid_type == current_pidrelations.primary_pid_type,
        ).one_or_none()
    relations = None
    if pid:
        relations = serialize_relations(pid)
        if relations:
            json['relations'] = relations
    return json



# -*- coding: utf-8 -*-
#
# This file is part of Invenio.
# Copyright (C) 2017 CERN.
#
# Invenio is free software; you can redistribute it
# and/or modify it under the terms of the GNU General Public License as
# published by the Free Software Foundation; either version 2 of the
# License, or (at your option) any later version.
#
# Invenio is distributed in the hope that it will be
# useful, but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the GNU
# General Public License for more details.
#
# You should have received a copy of the GNU General Public License
# along with Invenio; if not, write to the
# Free Software Foundation, Inc., 59 Temple Place, Suite 330, Boston,
# MA 02111-1307, USA.
#
# In applying this license, CERN does not
# waive the privileges and immunities granted to it by virtue of its status
# as an Intergovernmental Organization or submit itself to any jurisdiction.

"""PIDRelation JSON Schema for metadata."""

from marshmallow import Schema, fields

from ..utils import resolve_relation_type_config
from .utils import serialize_relations


class PIDSchema(Schema):
    """PID schema."""

    pid_type = fields.String()
    pid_value = fields.String()


class RelationSchema(Schema):
    """Generic PID relation schema."""

    # NOTE: Maybe do `fields.Function` for all of these and put them in `utils`
    parent = fields.Method('dump_parent')
    children = fields.Method('dump_children')
    type = fields.Method('dump_type')
    is_ordered = fields.Boolean()
    is_parent = fields.Method('_is_parent')
    is_child = fields.Method('_is_child')
    is_last = fields.Method('dump_is_last')
    index = fields.Method('dump_index')
    next = fields.Method('dump_next')
    previous = fields.Method('dump_previous')

    def _dump_relative(self, relative):
        if relative:
            data, errors = PIDSchema().dump(relative)
            return data
        else:
            return None

    def dump_next(self, obj):
        """Dump the parent of a PID."""
        if self._is_child(obj):
            return self._dump_relative(obj.next)

    def dump_previous(self, obj):
        """Dump the parent of a PID."""
        if self._is_child(obj):
            return self._dump_relative(obj.previous)

    def dump_index(self, obj):
        """Dump the index of the child in the relation."""
        if obj.is_ordered and self._is_child(obj):
            return obj.index
        else:
            return None

    def _is_parent(self, obj):
        """Check if the PID from the context is the parent in the relation."""
        return obj.parent == self.context['pid']

    def _is_child(self, obj):
        """Check if the PID from the context is the child in the relation."""
        return obj.child == self.context['pid']

    def dump_is_last(self, obj):
        """Dump the boolean stating if the child in the relation is last.

        Dumps `None` for parent serialization.
        """
        if self._is_child(obj) and obj.is_ordered:
            if obj.children.count() > 0:
                return obj.children.all()[-1] == self.context['pid']
            elif obj.draft_child:
                return obj.draft_child == self.context['pid']
            else:
                return True
        else:
            return None

    def dump_type(self, obj):
        """Dump the text name of the relation."""
        return resolve_relation_type_config(obj.relation_type).name

    def dump_parent(self, obj):
        """Dump the parent of a PID."""
        return self._dump_relative(obj.parent)

    def dump_children(self, obj):
        """Dump the siblings of a PID."""
        data, errors = PIDSchema(many=True).dump(obj.children.all())
        return data


class PIDRelationsMixin(object):
    """Mixin for easy inclusion of relations information in Record schemas."""

    relations = fields.Method('dump_relations')

    def dump_relations(self, obj):
        """Dump the relations to a dictionary."""
        pid = self.context['pid']
        return serialize_relations(pid)



# -*- coding: utf-8 -*-
#
# This file is part of Invenio.
# Copyright (C) 2017 CERN.
#
# Invenio is free software; you can redistribute it
# and/or modify it under the terms of the GNU General Public License as
# published by the Free Software Foundation; either version 2 of the
# License, or (at your option) any later version.
#
# Invenio is distributed in the hope that it will be
# useful, but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the GNU
# General Public License for more details.
#
# You should have received a copy of the GNU General Public License
# along with Invenio; if not, write to the
# Free Software Foundation, Inc., 59 Temple Place, Suite 330, Boston,
# MA 02111-1307, USA.
#
# In applying this license, CERN does not
# waive the privileges and immunities granted to it by virtue of its status
# as an Intergovernmental Organization or submit itself to any jurisdiction.

"""Invenio module that adds PID relations to the Invenio-PIDStore module."""

from collections import namedtuple

RelationType = namedtuple('RelationType',
                          ['id', 'name', 'label', 'api', 'schema'])

PIDRELATIONS_RELATION_TYPES = [
    RelationType(0, 'ordered', 'Ordered',
                 'invenio_pidrelations.api:PIDConceptOrdered',
                 'invenio_pidrelations.serializers.schemas.RelationSchema'),
    RelationType(1, 'unordered', 'Unordered',
                 'invenio_pidrelations.api:PIDConcept',
                 'invenio_pidrelations.serializers.schemas.RelationSchema'),
    RelationType(2, 'version', 'Version',
                 'invenio_pidrelations.contrib.versioning:PIDVersioning',
                 'invenio_pidrelations.serializers.schemas.RelationSchema'),
    RelationType(3, 'record_draft', 'Record Draft',
                 'invenio_pidrelations.contrib.records:RecordDraft',
                 'invenio_pidrelations.serializers.schemas.RelationSchema'),
]


PIDRELATIONS_PRIMARY_PID_TYPE = 'recid'
"""Default PID type for relations."""

PIDRELATIONS_INDEX_RELATIONS = False
"""Enable or disable relations indexing."""

PIDRELATIONS_DEFAULT_VALUE = 'foobar'
"""Default value for the application."""

PIDRELATIONS_INDEXED_RELATIONS = dict(
    recid=dict(
        field='version',
        api='invenio_pidrelations.contrib.versioning:PIDVersioning',
        # FIXME: for now the API does not provide any way to know if a relation
        # is ordered or not. Thus we write it here.
        ordered=True,
    )
)
"""Default PID fetcher."""



# -*- coding: utf-8 -*-
#
# This file is part of Invenio.
# Copyright (C) 2014, 2015, 2016 CERN.
#
# Invenio is free software; you can redistribute it
# and/or modify it under the terms of the GNU General Public License as
# published by the Free Software Foundation; either version 2 of the
# License, or (at your option) any later version.
#
# Invenio is distributed in the hope that it will be
# useful, but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the GNU
# General Public License for more details.
#
# You should have received a copy of the GNU General Public License
# along with Invenio; if not, write to the
# Free Software Foundation, Inc., 59 Temple Place, Suite 330, Boston,
# MA 02111-1307, USA.
#
# In applying this license, CERN does not
# waive the privileges and immunities granted to it by virtue of its status
# as an Intergovernmental Organization or submit itself to any jurisdiction.

"""API for PID relations concepts."""

from __future__ import absolute_import, print_function

from flask import Blueprint
from invenio_db import db
from invenio_pidstore.models import PersistentIdentifier, PIDStatus

from ..api import PIDConceptOrdered
from ..models import PIDRelation
from ..utils import resolve_relation_type_config


class PIDVersioning(PIDConceptOrdered):
    """API for PID versioning relations.

    - Adds automatic redirection handling for Parent-LastChild
    - Sets stricter method signatures, e.g.: 'index' is mandatory parameter
        when calling 'insert'.
    """

    def __init__(self, child=None, parent=None, draft_deposit=None,
                 draft_record=None, relation=None):
        """Create a PID versioning API."""
        self.relation_type = resolve_relation_type_config('version').id
        if relation is not None:
            if relation.relation_type != self.relation_type:
                raise ValueError("Provided PID relation ({0}) is not a "
                                 "version relation.".format(relation))
            super(PIDVersioning, self).__init__(relation=relation)
        else:
            super(PIDVersioning, self).__init__(
                child=child, parent=parent, relation_type=self.relation_type,
                relation=relation)
        if self.child:
            self.relation = PIDRelation.query.filter(
                PIDRelation.child_id == self.child.id,
                PIDRelation.relation_type == self.relation_type,
            ).one_or_none()

    def insert_child(self, child, index=-1):
        """Insert child into versioning scheme.

        Parameter 'index' is has to be an integer.
        """
        if index is None:
            raise ValueError(
                "Incorrect value for child index: {0}".format(index))

        with db.session.begin_nested():
            super(PIDVersioning, self).insert_child(child, index=index)
            self.parent.redirect(child)

    def remove_child(self, child):
        """Remove a child from a versioning scheme.

        Extends the base method call with always reordering after removal and
        adding a redirection from the parent to the last child.
        """
        with db.session.begin_nested():
            super(PIDVersioning, self).remove_child(child, reorder=True)
            if self.last_child is not None:
                self.parent.redirect(self.last_child)

    def create_parent(self, pid_value, status=PIDStatus.REGISTERED,
                      redirect=True):
        """Create a parent PID from a child and create a new PID versioning."""
        if self.has_parents:
            raise Exception("Parent already exists for this child.")
        self.parent = PersistentIdentifier.create(
            self.child.pid_type, pid_value,
            object_type=self.child.object_type,
            status=status)
        self.relation = PIDRelation.create(
            self.parent, self.child, self.relation_type, 0)
        if redirect:
            self.parent.redirect(self.child)

    @property
    def exists(self):
        """Check if the PID Versioning exists."""
        return self.parent is not None

    @property
    def last_child(self):
        """
        Get the latest PID as pointed by the Head PID.

        If the 'pid' is a Head PID, return the latest of its children.
        If the 'pid' is a Version PID, return the latest of its siblings.
        Return None for the non-versioned PIDs.
        """
        return self.get_children(ordered=False,
                                 pid_status=PIDStatus.REGISTERED).filter(
                PIDRelation.index.isnot(None)).order_by(
                    PIDRelation.index.desc()).first()

    @property
    def draft_child(self):
        """Get the last non-registered child."""
        return self.get_children(ordered=False).filter(
                PIDRelation.index.isnot(None),
                PersistentIdentifier.status == PIDStatus.RESERVED).order_by(
                    PIDRelation.index.desc()).one_or_none()

    @property
    def draft_child_deposit(self):
        """
        Get the deposit of the draft child.

        Return `None` if no new-version deposit exists.
        """
        from invenio_pidrelations.contrib.records import RecordDraft
        if self.draft_child:
            return RecordDraft.get_draft(self.draft_child)
        else:
            return None

    def insert_draft_child(self, child):
        """Insert a draft child to versioning."""
        if not self.draft_child:
            with db.session.begin_nested():
                super(PIDVersioning, self).insert_child(child, index=-1)
        else:
            raise Exception(
                "Draft child already exists for this relation: {0}".format(
                    self.draft_child))

    def remove_draft_child(self):
        """Remove the draft child from versioning."""
        if self.draft_child:
            with db.session.begin_nested():
                super(PIDVersioning, self).remove_child(self.draft_child,
                                                        reorder=True)

    def update_redirect(self):
        """Update the parent redirect to the current last child."""
        if self.last_child:
            if self.parent.status == PIDStatus.RESERVED:
                self.parent.register()
            self.parent.redirect(self.last_child)

    @property
    def children(self):
        """Children of the parent."""
        return self.get_children(pid_status=PIDStatus.REGISTERED, ordered=True)


versioning_blueprint = Blueprint(
    'invenio_pidrelations.versioning',
    __name__,
    template_folder='templates'
)


@versioning_blueprint.app_template_filter()
def to_versioning_api(pid, child=True):
    """Get PIDVersioning object."""
    return PIDVersioning(
        child=pid if child else None,
        parent=pid if not child else None
    )


__all__ = (
    'PIDVersioning',
    'versioning_blueprint'
)



# -*- coding: utf-8 -*-
#
# This file is part of Invenio.
# Copyright (C) 2017 CERN.
#
# Invenio is free software; you can redistribute it
# and/or modify it under the terms of the GNU General Public License as
# published by the Free Software Foundation; either version 2 of the
# License, or (at your option) any later version.
#
# Invenio is distributed in the hope that it will be
# useful, but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the GNU
# General Public License for more details.
#
# You should have received a copy of the GNU General Public License
# along with Invenio; if not, write to the
# Free Software Foundation, Inc., 59 Temple Place, Suite 330, Boston,
# MA 02111-1307, USA.
#
# In applying this license, CERN does not
# waive the privileges and immunities granted to it by virtue of its status
# as an Intergovernmental Organization or submit itself to any jurisdiction.

"""Module tests."""

from __future__ import absolute_import, print_function

from invenio_pidstore.models import PersistentIdentifier, PIDStatus

from invenio_pidrelations.contrib.versioning import PIDVersioning
from invenio_pidrelations.models import PIDRelation
from invenio_pidrelations.utils import resolve_relation_type_config


def test_version_pids_create(app, db):

    # Create a child, initialize the Versioning API and create a parent
    assert PersistentIdentifier.query.count() == 0
    # Create a child
    h1v1 = PersistentIdentifier.create('recid', '12345', object_type='rec',
                                       status=PIDStatus.REGISTERED)
    assert PersistentIdentifier.query.count() == 1
    pv = PIDVersioning(child=h1v1)
    # Create a parent
    pv.create_parent('12345.parent')
    assert PersistentIdentifier.query.count() == 2
    assert pv.parent.get_redirect() == h1v1
    assert pv.parent.status == PIDStatus.REDIRECTED
    # Make sure 'pid_type', 'object_type' and 'status' are inherited from child
    assert pv.parent.pid_type == pv.child.pid_type
    assert pv.parent.object_type == pv.child.object_type

    pr = PIDRelation.query.one()
    assert pr.child == h1v1
    assert pr.parent == pv.parent

    VERSION = resolve_relation_type_config('version').id
    assert pr.relation_type == VERSION
    assert pr.index == 0


def test_version_api_edit(app, db, version_pids):
    h1, h1v1, h1v2 = (version_pids[p] for p in ['h1', 'h1v1', 'h1v2'])
    pv = PIDVersioning(parent=h1)

    assert [h1v1, h1v2] == pv.children.all()
    assert h1.get_redirect() == h1v2
    assert pv.last_child == h1v2

    h1v3 = PersistentIdentifier.create('recid', 'foobar.v3', object_type='rec',
                                       status=PIDStatus.REGISTERED)
    pv.insert_child(h1v3)
    assert [h1v1, h1v2, h1v3] == pv.children.all()
    assert h1.get_redirect() == h1v3
    pv.last_child == h1v3

    pv.remove_child(h1v3)
    assert h1.get_redirect() == h1v2
    assert pv.last_child == h1v2

    pv.remove_child(h1v2)
    assert h1.get_redirect() == h1v1
    assert pv.last_child == h1v1



# -*- coding: utf-8 -*-
#
# This file is part of Invenio.
# Copyright (C) 2017 CERN.
#
# Invenio is free software; you can redistribute it
# and/or modify it under the terms of the GNU General Public License as
# published by the Free Software Foundation; either version 2 of the
# License, or (at your option) any later version.
#
# Invenio is distributed in the hope that it will be
# useful, but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the GNU
# General Public License for more details.
#
# You should have received a copy of the GNU General Public License
# along with Invenio; if not, write to the
# Free Software Foundation, Inc., 59 Temple Place, Suite 330, Boston,
# MA 02111-1307, USA.
#
# In applying this license, CERN does not
# waive the privileges and immunities granted to it by virtue of its status
# as an Intergovernmental Organization or submit itself to any jurisdiction.

"""PIDRelation serialization utilities."""

from invenio_pidrelations.api import PIDRelation

from ..utils import resolve_relation_type_config


def serialize_relations(pid):
    """Serialize the relations for given PID."""
    data = {}
    relations = PIDRelation.get_child_relations(pid).all()
    parent_relation = PIDRelation.get_parent_relations(pid).first()
    if parent_relation:
        relations.append(parent_relation)

    for relation in relations:
        rel_cfg = resolve_relation_type_config(relation.relation_type)
        schema_class = rel_cfg.schema
        if schema_class is not None:
            schema = schema_class()
            schema.context['pid'] = pid
            result, errors = schema.dump(rel_cfg.api(relation=relation))
            data.setdefault(rel_cfg.name, []).append(result)
    return data



# -*- coding: utf-8 -*-
#
# This file is part of Invenio.
# Copyright (C) 2017 CERN.
#
# Invenio is free software; you can redistribute it
# and/or modify it under the terms of the GNU General Public License as
# published by the Free Software Foundation; either version 2 of the
# License, or (at your option) any later version.
#
# Invenio is distributed in the hope that it will be
# useful, but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the GNU
# General Public License for more details.
#
# You should have received a copy of the GNU General Public License
# along with Invenio; if not, write to the
# Free Software Foundation, Inc., 59 Temple Place, Suite 330, Boston,
# MA 02111-1307, USA.
#
# In applying this license, CERN does not
# waive the privileges and immunities granted to it by virtue of its status
# as an Intergovernmental Organization or submit itself to any jurisdiction.

"""Invenio module that adds PID relations to the Invenio-PIDStore module."""

from __future__ import absolute_import, print_function

from copy import deepcopy

from invenio_indexer.signals import before_record_index
from werkzeug.utils import cached_property

from invenio_pidrelations import config
from invenio_pidrelations.indexers import index_relations
from invenio_pidrelations.utils import obj_or_import_string


class _InvenioPIDRelationsState(object):
    """InvenioPIDRelations REST state."""

    def __init__(self, app):
        """Initialize state."""
        self.app = app

    @cached_property
    def relation_types(self):
        return self.app.config.get('PIDRELATIONS_RELATION_TYPES', {})

    @cached_property
    def primary_pid_type(self):
        return self.app.config.get('PIDRELATIONS_PRIMARY_PID_TYPE')

    def indexed_relations(self):
        """Load the configuration for indexed relations."""
        indexed = self.app.config.get('PIDRELATIONS_INDEXED_RELATIONS')
        if not indexed:
            return {}
        result = deepcopy(indexed)
        for pid_value, conf in result.items():
            conf.update(dict(api=obj_or_import_string(conf['api'])))
        return result


class InvenioPIDRelations(object):
    """Invenio-PIDRelations extension."""

    def __init__(self, app=None):
        """Extension initialization."""
        if app:
            self.init_app(app)

    def init_app(self, app):
        """Flask application initialization."""
        self.init_config(app)
        app.extensions['invenio-pidrelations'] = _InvenioPIDRelationsState(app)

        # Register indexers the default indexer if necessary
        if app.config.get('PIDRELATIONS_REGISTER_DEFAULT_INDEXER'):
            before_record_index.connect(index_relations, sender=app)

    def init_config(self, app):
        """Initialize configuration."""
        for k in dir(config):
            if k.startswith('PIDRELATIONS_'):
                app.config.setdefault(k, getattr(config, k))



# -*- coding: utf-8 -*-
#
# This file is part of Invenio.
# Copyright (C) 2017 CERN.
#
# Invenio is free software; you can redistribute it
# and/or modify it under the terms of the GNU General Public License as
# published by the Free Software Foundation; either version 2 of the
# License, or (at your option) any later version.
#
# Invenio is distributed in the hope that it will be
# useful, but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the GNU
# General Public License for more details.
#
# You should have received a copy of the GNU General Public License
# along with Invenio; if not, write to the
# Free Software Foundation, Inc., 59 Temple Place, Suite 330, Boston,
# MA 02111-1307, USA.
#
# In applying this license, CERN does not
# waive the privileges and immunities granted to it by virtue of its status
# as an Intergovernmental Organization or submit itself to any jurisdiction.

"""Module tests."""

from __future__ import absolute_import, print_function

from invenio_pidrelations.api import PIDConcept
from invenio_pidrelations.models import PIDRelation
from invenio_pidrelations.utils import resolve_relation_type_config


def test_basic_api_methods(app, db, pids):
    # Set-up
    h1, h1v1, h1v2, h1v3, h2, h2v1, c1, c1r1, c1r2, pid1 = \
        (pids[p] for p in ['h1', 'h1v1', 'h1v2', 'h1v3', 'h2', 'h2v1',
                           'c1', 'c1r1', 'c1r2', 'pid1'])
    ORDERED = resolve_relation_type_config('ordered').id
    UNORDERED = resolve_relation_type_config('unordered').id

    # Test the "children" property and "get_children" method
    # Ordered relations
    assert PIDConcept(
        parent=h1, relation_type=ORDERED).children.count() == 3
    assert set([h1v3, h1v2, h1v1]) == set(PIDConcept(
        parent=h1, relation_type=ORDERED).children.all())
    assert [h1v1, h1v2, h1v3] == PIDConcept(
        parent=h1, relation_type=ORDERED).get_children(ordered=True).all()
    assert PIDConcept(
        parent=h1v1, relation_type=ORDERED).children.count() == 0
    # Unordered relations
    assert PIDConcept(
        parent=c1, relation_type=UNORDERED).children.count() == 2
    assert set([c1r1, c1r2]) == set(PIDConcept(
        parent=c1, relation_type=UNORDERED).children.all())
    assert PIDConcept(
        parent=c1r1, relation_type=UNORDERED).children.count() == 0
    # Unrelated PIDs
    assert PIDConcept(
        parent=pid1, relation_type=ORDERED).children.count() == 0

    # Test the "parents" property
    assert PIDConcept(child=h1, relation_type=ORDERED).parents.count() == 0
    assert h1 == PIDConcept(child=h1v1, relation_type=ORDERED).parents.one()
    assert h1 == PIDConcept(child=h1v2, relation_type=ORDERED).parents.one()

    # # Test the "parent" property
    assert h1 == PIDConcept(child=h1v1, relation_type=ORDERED).parent
    assert h1 == PIDConcept(child=h1v2, relation_type=ORDERED).parent

    # Test "has_children", "has_parents"
    # Ordered relations
    assert PIDConcept(parent=h1, relation_type=ORDERED).has_children is True
    assert PIDConcept(child=h1, relation_type=ORDERED).has_parents is False

    assert PIDConcept(
        parent=h1v1, relation_type=ORDERED).has_children is False
    assert PIDConcept(
        child=h1v1, relation_type=ORDERED).has_parents is True

    # Unordered relations
    assert PIDConcept(
        parent=c1, relation_type=UNORDERED).has_children is True
    assert PIDConcept(
        child=c1, relation_type=UNORDERED).has_parents is False

    # Unrelated PIDs
    assert PIDConcept(
        parent=pid1, relation_type=ORDERED).has_children is False
    assert PIDConcept(
        child=pid1, relation_type=ORDERED).has_parents is False

    # Test "is_child", "is_parent" properties
    assert PIDConcept(parent=h1, relation_type=ORDERED).is_parent is True
    assert PIDConcept(parent=h1v1, relation_type=ORDERED).is_parent is False
    assert PIDConcept(child=h1v1, relation_type=ORDERED).is_child is True
    assert PIDConcept(child=h1, relation_type=ORDERED).is_child is False
    pidc1 = PIDConcept(parent=h1, child=h1v1, relation_type=ORDERED)
    assert pidc1.is_child is True
    assert pidc1.is_parent is True

    assert PIDConcept(parent=pid1, relation_type=ORDERED).is_parent is False
    assert PIDConcept(child=pid1, relation_type=ORDERED).is_child is False
    pidc1 = PIDConcept(parent=pid1, child=pid1, relation_type=ORDERED)
    assert pidc1.is_child is False
    assert pidc1.is_parent is False

    # Test "last_child", "is_last_child" properties
    # Ordered relations
    assert h1v3 == PIDConcept(parent=h1, relation_type=ORDERED).last_child
    assert PIDConcept(child=h1v3, relation_type=ORDERED).is_last_child is True
    assert PIDConcept(child=h1v2, relation_type=ORDERED).is_last_child is False
    assert PIDConcept(child=h1v1, relation_type=ORDERED).is_last_child is False

    # Unordered relations
    assert PIDConcept(parent=c1, relation_type=UNORDERED).last_child is None
    assert PIDConcept(parent=c1r1,
                      relation_type=UNORDERED).is_last_child is False
    assert PIDConcept(parent=c1r2,
                      relation_type=UNORDERED).is_last_child is False

    # Unrelated PIDs
    assert PIDConcept(parent=pid1, relation_type=ORDERED).last_child is None
    assert PIDConcept(parent=pid1,
                      relation_type=ORDERED).is_last_child is False
    assert PIDConcept(parent=pid1, relation_type=UNORDERED).last_child is None
    assert PIDConcept(parent=pid1,
                      relation_type=UNORDERED).is_last_child is False

    # Test the "insert" method for ordered relations
    # Insert at the end (index=-1)
    PIDConcept(parent=h1, relation_type=ORDERED).insert_child(pid1, index=-1)
    assert [h1v1, h1v2, h1v3, pid1] == PIDConcept(
        parent=h1, relation_type=ORDERED).get_children(ordered=True).all()
    # Make sure relations order is preserved
    h1_c = h1.child_relations.order_by(PIDRelation.index).all()
    assert h1_c[0].child == h1v1
    assert h1_c[0].index == 0
    assert h1_c[1].child == h1v2
    assert h1_c[1].index == 1
    assert h1_c[2].child == h1v3
    assert h1_c[2].index == 2
    assert h1_c[3].child == pid1
    assert h1_c[3].index == 3

    # Return to previous state
    PIDConcept(parent=h1, relation_type=ORDERED).remove_child(
        pid1, reorder=True)
    assert [h1v1, h1v2, h1v3] == PIDConcept(
        parent=h1, relation_type=ORDERED).get_children(ordered=True).all()
    h1_c = h1.child_relations.order_by(PIDRelation.index).all()
    assert h1_c[0].child == h1v1
    assert h1_c[0].index == 0
    assert h1_c[1].child == h1v2
    assert h1_c[1].index == 1

    # Insert at the first position (index=0)
    PIDConcept(parent=h1, relation_type=ORDERED).insert_child(pid1, index=0)
    assert [pid1, h1v1, h1v2, h1v3] == PIDConcept(
        parent=h1, relation_type=ORDERED).get_children(ordered=True).all()
    h1_c = h1.child_relations.order_by(PIDRelation.index).all()
    assert h1_c[0].child == pid1
    assert h1_c[0].index == 0
    assert h1_c[1].child == h1v1
    assert h1_c[1].index == 1
    assert h1_c[2].child == h1v2
    assert h1_c[2].index == 2
    PIDConcept(parent=h1, relation_type=ORDERED).remove_child(
        pid1, reorder=True)
    assert [h1v1, h1v2, h1v3] == PIDConcept(
        parent=h1, relation_type=ORDERED).get_children(ordered=True).all()
    h1_c = h1.child_relations.order_by(PIDRelation.index).all()
    assert h1_c[0].child == h1v1
    assert h1_c[0].index == 0
    assert h1_c[1].child == h1v2
    assert h1_c[1].index == 1

    # Insert at the second position (index=1)
    PIDConcept(parent=h1, relation_type=ORDERED).insert_child(pid1, index=1)
    assert [h1v1, pid1, h1v2, h1v3] == PIDConcept(
        parent=h1, relation_type=ORDERED).get_children(ordered=True).all()
    h1_c = h1.child_relations.order_by(PIDRelation.index).all()
    assert h1_c[0].child == h1v1
    assert h1_c[0].index == 0
    assert h1_c[1].child == pid1
    assert h1_c[1].index == 1
    assert h1_c[2].child == h1v2
    assert h1_c[2].index == 2
    PIDConcept(parent=h1, relation_type=ORDERED).remove_child(
        pid1, reorder=True)
    assert [h1v1, h1v2, h1v3] == PIDConcept(
        parent=h1, relation_type=ORDERED).get_children(ordered=True).all()
    h1_c = h1.child_relations.order_by(PIDRelation.index).all()
    assert h1_c[0].child == h1v1
    assert h1_c[0].index == 0
    assert h1_c[1].child == h1v2
    assert h1_c[1].index == 1

    # Insert at an arbitrarily large position (appends to the end)
    PIDConcept(parent=h1, relation_type=ORDERED).insert_child(pid1, index=100)
    assert [h1v1, h1v2, h1v3, pid1] == PIDConcept(
        parent=h1, relation_type=ORDERED).get_children(ordered=True).all()
    h1_c = h1.child_relations.order_by(PIDRelation.index).all()
    assert h1_c[0].child == h1v1
    assert h1_c[0].index == 0
    assert h1_c[1].child == h1v2
    assert h1_c[1].index == 1
    assert h1_c[2].child == h1v3
    assert h1_c[2].index == 2
    assert h1_c[3].child == pid1
    assert h1_c[3].index == 3
    PIDConcept(parent=h1, relation_type=ORDERED).remove_child(
        pid1, reorder=True)
    assert [h1v1, h1v2, h1v3] == PIDConcept(
        parent=h1, relation_type=ORDERED).get_children(ordered=True).all()
    h1_c = h1.child_relations.order_by(PIDRelation.index).all()
    assert h1_c[0].child == h1v1
    assert h1_c[0].index == 0
    assert h1_c[1].child == h1v2
    assert h1_c[1].index == 1

    # Remove an extra child
    PIDConcept(parent=h1, relation_type=ORDERED).remove_child(
        h1v1, reorder=True)
    assert [h1v2, h1v3] == PIDConcept(
        parent=h1, relation_type=ORDERED).get_children(ordered=True).all()
    h1_c = h1.child_relations.order_by(PIDRelation.index).all()
    assert h1.child_relations.count() == 2
    assert h1_c[0].child == h1v2
    assert h1_c[0].index == 0



# -*- coding: utf-8 -*-
#
# This file is part of Invenio.
# Copyright (C) 2014, 2015, 2016 CERN.
#
# Invenio is free software; you can redistribute it
# and/or modify it under the terms of the GNU General Public License as
# published by the Free Software Foundation; either version 2 of the
# License, or (at your option) any later version.
#
# Invenio is distributed in the hope that it will be
# useful, but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the GNU
# General Public License for more details.
#
# You should have received a copy of the GNU General Public License
# along with Invenio; if not, write to the
# Free Software Foundation, Inc., 59 Temple Place, Suite 330, Boston,
# MA 02111-1307, USA.
#
# In applying this license, CERN does not
# waive the privileges and immunities granted to it by virtue of its status
# as an Intergovernmental Organization or submit itself to any jurisdiction.

"""API for PID relations concepts."""

from __future__ import absolute_import, print_function

from invenio_db import db
from invenio_pidstore.models import PersistentIdentifier
from sqlalchemy.exc import IntegrityError

from .models import PIDRelation


class PIDConcept(object):
    """API for PID version relations."""

    def __init__(self, child=None, parent=None, relation_type=None,
                 relation=None):
        """Create a PID concept API object."""
        if relation:
            self.relation = relation
            self.child = relation.child
            self.parent = relation.parent
            self.relation_type = relation.relation_type
        else:
            self.child = child
            self.parent = parent
            self.relation_type = relation_type
            # If child and parent (primary keys) are not None,
            # try to set the relation
            if child and parent:
                self.relation = PIDRelation.query.get(
                    (self.parent.id, self.child.id))
            else:
                self.relation = None

    @property
    def parents(self):
        """Return the PID parents for given relation."""
        filter_cond = [PIDRelation.child_id == self.child.id, ]
        if self.relation_type is not None:
            filter_cond.append(PIDRelation.relation_type == self.relation_type)
        return db.session.query(PersistentIdentifier).join(
            PIDRelation,
            PIDRelation.parent_id == PersistentIdentifier.id
        ).filter(*filter_cond)

    @property
    def exists(self):
        """Determine if a PID Concept exists.

        Determine if constructed API object describes an existing PID Concept.
        The definition of that will vary across different PID Concepts, but
        it's intended use is to check if given child/parent PIDs are in
        the described relation.
        """
        return bool(self.relation)

    @property
    def is_ordered(self):
        """Determine if the concept is an ordered concept."""
        return all(val is not None for val in self.children.with_entities(
            PIDRelation.index))

    @property
    def has_parents(self):
        """Determine if there are any parents in this relationship."""
        return self.parents.count() > 0

    @property
    def parent(self):
        """Return the parent of the PID in given relation.

        NOTE: Not supporting relations, which allow for multiple parents,
              e.g. Collection.

        None if not found
        Raises 'sqlalchemy.orm.exc.MultipleResultsFound' for multiple parents.
        """
        if self._parent is None:
            parent = self.parents.one_or_none()
            self._parent = parent
        return self._parent

    @parent.setter
    def parent(self, parent):
        self._parent = parent

    @property
    def is_parent(self):
        """Determine if the provided parent is a parent in the relation."""
        return self.has_children

    def get_children(self, ordered=False, pid_status=None):
        """Get all children of the parent."""
        filter_cond = [PIDRelation.parent_id == self.parent.id, ]
        if pid_status is not None:
            filter_cond.append(PersistentIdentifier.status == pid_status)
        if self.relation_type is not None:
            filter_cond.append(PIDRelation.relation_type == self.relation_type)

        q = db.session.query(PersistentIdentifier).join(
            PIDRelation,
            PIDRelation.child_id == PersistentIdentifier.id
        ).filter(*filter_cond)
        if ordered:
            return q.order_by(PIDRelation.index.asc())
        else:
            return q

    @property
    def index(self):
        """Index of the child in the relation."""
        return self.relation.index

    @property
    def children(self):
        """Children of the parent."""
        return self.get_children()

    @property
    def has_children(self):
        """Determine if there are any children in this relationship."""
        return self.children.count() > 0

    @property
    def is_last_child(self):
        """
        Determine if 'pid' is the latest version of a resource.

        Resolves True for Versioned PIDs which are the oldest of its siblings.
        False otherwise, also for Head PIDs.
        """
        last_child = self.last_child
        if last_child is None:
            return False
        return last_child == self.child

    @property
    def last_child(self):
        """
        Get the latest PID as pointed by the Head PID.

        If the 'pid' is a Head PID, return the latest of its children.
        If the 'pid' is a Version PID, return the latest of its siblings.
        Return None for the non-versioned PIDs.
        """
        return self.get_children(ordered=False).filter(
            PIDRelation.index.isnot(None)).order_by(
                PIDRelation.index.desc()).first()

    @property
    def next(self):
        """Get the next sibling in the PID relation."""
        if self.relation.index is not None:
            return self.children.filter_by(
                index=self.relation.index + 1).one_or_none()
        else:
            return None

    @property
    def previous(self):
        """Get the previous sibling in the PID relation."""
        if self.relation.index is not None:
            return self.children.filter_by(
                index=self.relation.index - 1).one_or_none()
        else:
            return None

    @property
    def is_child(self):
        """
        Determine if 'pid' is a Version PID.

        Resolves as True for any PID which has a Head PID, False otherwise.
        """
        return self.has_parents

    def insert_child(self, child, index=None):
        """Insert a new child into a PID concept.

        Argument 'index' can take the following values:
            0,1,2,... - insert child PID at the specified position
            -1 - insert the child PID at the last position
            None - insert child without order (no re-ordering is done)

            NOTE: If 'index' is specified, all sibling relations should
                  have PIDRelation.index information.

        """
        try:
            with db.session.begin_nested():
                if index is not None:
                    child_relations = self.parent.child_relations.filter(
                        PIDRelation.relation_type ==
                        self.relation_type).order_by(PIDRelation.index).all()
                    relation_obj = PIDRelation.create(
                        self.parent, child, self.relation_type, None)
                    if index == -1:
                        child_relations.append(relation_obj)
                    else:
                        child_relations.insert(index, relation_obj)
                    for idx, c in enumerate(child_relations):
                        c.index = idx
                else:
                    relation_obj = PIDRelation.create(
                        self.parent, child, self.relation_type, None)
        except IntegrityError:
            raise Exception("PID Relation already exists.")

    def remove_child(self, child, reorder=False):
        """Remove a child from a PID concept."""
        with db.session.begin_nested():
            relation = PIDRelation.query.filter_by(
                parent_id=self.parent.id,
                child_id=child.id,
                relation_type=self.relation_type).one()
            db.session.delete(relation)
            if reorder:
                child_relations = self.parent.child_relations.filter(
                    PIDRelation.relation_type == self.relation_type).order_by(
                        PIDRelation.index).all()
                for idx, c in enumerate(child_relations):
                    c.index = idx


class PIDConceptOrdered(PIDConcept):
    """Standard PID Concept with childred ordering."""

    @property
    def children(self):
        """Overwrite the children property to always return them ordered."""
        return self.get_children(ordered=True)

    @property
    def is_ordered(self):
        """Determine if the concept is an ordered concept."""
        return True


__all__ = (
    'PIDConcept',
    'PIDConceptOrdered',
)



# -*- coding: utf-8 -*-
#
# This file is part of Invenio.
# Copyright (C) 2017 CERN.
#
# Invenio is free software; you can redistribute it
# and/or modify it under the terms of the GNU General Public License as
# published by the Free Software Foundation; either version 2 of the
# License, or (at your option) any later version.
#
# Invenio is distributed in the hope that it will be
# useful, but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the GNU
# General Public License for more details.
#
# You should have received a copy of the GNU General Public License
# along with Invenio; if not, write to the
# Free Software Foundation, Inc., 59 Temple Place, Suite 330, Boston,
# MA 02111-1307, USA.
#
# In applying this license, CERN does not
# waive the privileges and immunities granted to it by virtue of its status
# as an Intergovernmental Organization or submit itself to any jurisdiction.

"""Records integration for PIDRelations."""

from invenio_indexer.api import RecordIndexer
from invenio_pidstore.models import PersistentIdentifier
from invenio_records.api import Record

from ..api import PIDConcept
from ..contrib.versioning import PIDVersioning
from ..utils import resolve_relation_type_config


class RecordDraft(PIDConcept):
    """Record Draft relationship.

    Users of this class should make calls to `link` and `unlink` based on their
    specific use-cases. Typical scenario is that of creating a new Deposit
    which is linked to a not-yet published record PID (PID status is RESERVED).
    Linking these two makes it possible for creating new record versions and
    having the abiltiy to track their deposits.

    NOTE: This relation exists because usually newly created records are not
    immediately stored inside the database (they have no `RecordMetada`). This
    leads to having deposits that are hanging onto no actual record and only
    possess "soft" links to their records' PIDs through metadata.
    """

    def __init__(self, child=None, parent=None, relation=None):
        """Create a record draft relation."""
        self.relation_type = resolve_relation_type_config('record_draft').id
        if relation is not None:
            if relation.relation_type != self.relation_type:
                raise ValueError('Provided PID relation ({0}) is not a '
                                 'version relation.'.format(relation))
            super(RecordDraft, self).__init__(relation=relation)
        else:
            # RecordDraft is composed of a one parent-child pair, hence we can
            # determine the child from the parent
            self.parent = parent
            self.child = child
            if parent:
                self.child = self.children.one_or_none()
            if child:
                self.parent = self.parents.one_or_none()
            # Run the parent class constructor in attempt to set the relation
            super(RecordDraft, self).__init__(
                child=self.child, parent=self.parent,
                relation_type=self.relation_type, relation=relation)

    @classmethod
    def link(cls, recid, depid):
        """Link a recid and depid."""
        recid_api = cls(parent=recid)
        depid_api = cls(child=depid)
        if recid_api.has_children:
            raise Exception('Recid {} already has a depid as a draft.'
                            .format(recid))
        if depid_api.parent:
            raise Exception('Depid {} already is a draft of a recid.'
                            .format(recid))
        recid_api.insert_child(depid)

    @classmethod
    def unlink(cls, recid, depid):
        """Unlink a recid and depid."""
        return cls(parent=recid).remove_child(depid)

    @classmethod
    def get_draft(cls, recid):
        """Get the draft of a record."""
        api = cls(parent=recid)
        if api.exists:
            return api.children.one_or_none()
        else:
            return None

    @classmethod
    def get_recid(cls, depid):
        """Get the recid of a record."""
        api = cls(child=depid)
        if api.exists:
            return api.parent
        else:
            return None


def get_latest_draft(recid_pid):
    """Return the latest draft for a record."""
    pv = PIDVersioning(child=recid_pid)
    if pv.draft_child:
        last_deposit = RecordDraft.get_draft(pv.draft_child)
    else:
        last_deposit = None
    return pv.last_child, last_deposit


def index_siblings(pid, children=None, neighbors_eager=False, eager=False,
                   with_deposits=True):
    """Send sibling records of the passed pid for indexing.

    Note: Does not index the 'pid' itself, only zero or more siblings.

    :param pid: PID (recid) of whose siblings are to be indexed.
    :param children: Overrides children with a fixed list of PID.
        Children should contain the 'pid' itself if 'neighbors_eager' is to
        be used, otherwise the last child is treated as the only neighbor.
    :param eager: Index all siblings immediately.
    :param neighbors_eager: Index the neighboring PIDs w.r.t. 'pid'
        immediately, and the rest with a bulk_index.
    :param with_deposits: Reindex also corresponding record's deposits.
    """
    assert not (neighbors_eager and eager), \
        "Only one of the 'eager' and 'neighbors_eager' flags can be set to " \
        "True, not both"
    if children is None:
        children = PIDVersioning(child=pid).children.all()

    objid = str(pid.object_uuid)
    children = [str(p.object_uuid) for p in children]

    idx = children.index(objid) if objid in children else len(children)

    # Split children (which can include the pid) into left and right siblings
    # If 'pid' is not in children, idx is the lenght of list, so 'left'
    # will be all children, and 'right' will be an empty list
    # [X X X] X [X X X]
    left = children[:idx]
    right = children[idx + 1:]

    if eager:
        eager_uuids = left + right
        bulk_uuids = []
    elif neighbors_eager:
        # neighbors are last of 'left' and first or 'right' siblings
        # X X [X] X [X] X X
        eager_uuids = left[-1:] + right[:1]
        # all of the siblings, except the neighbours
        # [X X] X X X [X X]
        bulk_uuids = left[:-1] + right[1:]
    else:
        eager_uuids = []
        bulk_uuids = left + right

    def get_dep_uuids(rec_uuids):
        """Get corresponding deposit UUIDs from record's UUIDs."""
        return [str(PersistentIdentifier.get('depid',
                    Record.get_record(id_)['_deposit']['id']).object_uuid)
                for id_ in rec_uuids]

    if with_deposits:
        eager_uuids += get_dep_uuids(eager_uuids)
        bulk_uuids += get_dep_uuids(bulk_uuids)

    for id_ in eager_uuids:
        RecordIndexer().index_by_id(id_)
    if bulk_uuids:
        RecordIndexer().bulk_index(bulk_uuids)
